import click
import pathlib
from serial.tools import list_ports
import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

drivers = entry_points(group='pysigrok.hardware')
driver_classes = {}
for hw in drivers:
    loaded = hw.load()
    driver_classes[loaded.name] = loaded

input_formats = entry_points(group='pysigrok.input_format')
input_classes = {}
for f in input_formats:
    loaded = f.load()
    input_classes[loaded.name] = loaded

output_formats = entry_points(group='pysigrok.output_format')
output_classes = {}
for f in output_formats:
    loaded = f.load()
    output_classes[loaded.name] = loaded

decoders = entry_points(group='pysigrok.decoders')
decoder_classes = {}
for decoder in decoders:
    loaded = decoder.load()
    decoder_classes[loaded.id] = loaded

@click.command()
@click.option("--list-supported", "-L", is_flag=True, default=False)
@click.option("--list-serial", is_flag=True, default=False)
@click.option("-d", "--driver")
@click.option("-c", "--config", "configs")
@click.option("-i", "--input-file")
@click.option("-I", "--input-format", default="srzip")
@click.option("-o", "--output-file")
@click.option("-O", "--output-format")
@click.option("-C", "--channels")
@click.option("--time", "sample_time")
@click.option("--samples", type=int)
@click.option("--frames")
@click.option("--continuous", is_flag=True)
def main(list_supported, list_serial, driver, configs, input_file, input_format, output_file, output_format, channels, sample_time, samples, frames, continuous):

    if list_supported:
        print("Supported hardware drivers:")
        for driver_id in driver_classes:
            driver_class = driver_classes[driver_id]
            print(f"  {driver_id}\t{driver_class.longname}")
        print()
        print("Supported input formats:")
        for input_format_id in input_classes:
            input_class = input_classes[input_format_id]
            print(f"  {input_format_id}\t{input_class.desc}")
        print()
        print("Supported output formats:")
        for output_format_id in output_classes:
            output_class = output_classes[output_format_id]
            print(f"  {output_format_id}\t{output_class.desc}")
        print()
        print("Supported transform modules:")
        print()
        print("Supported protocol decoders:")
        for pd in decoder_classes:
            decoder_class = decoder_classes[pd]
            print(f"  {pd}\t{decoder_class.longname}")
    elif list_serial:
        print("Available serial ports:")
        for port in list_ports.comports():
            print(" ", port)

    if driver:
        driver_options = {}
        if ":" in driver:
            driver, options = driver.split(":", maxsplit=1)
            for option in options.split(":"):
                k, v = option.split("=", maxsplit=1)
                driver_options[k] = v
        driver_configs = {}
        if configs:
            for config in configs.split(":"):
                k, v = config.split("=", maxsplit=1)
                driver_configs[k] = v

        driver_class = None
        for hw in drivers:
            loaded = hw.load()
            if loaded.name == driver:
                driver_class = loaded

        driver = driver_class(channels, **driver_options, **driver_configs)

        if samples:
            # acquire data
            driver.acquire(samples)

            if output_file:
                # Delete the file if it exists
                p = pathlib.Path(output_file)
                p.unlink(missing_ok=True)
                f = open(output_file, "wb")
                if not output_format:
                    output_format = "srzip"
            else:
                f = sys.stdout
                if not output_format:
                    output_format = "bits:width=64"
            output_options = {}
            if ":" in output_format:
                output_split = output_format.split(":")
                output_format_id = output_split[0]
                for option in output_split[1:]:
                    if "=" in option:
                        k, v = option.split("=")
                        output_options[k] = v
                    else:
                        output_options[option] = "true"
            else:
                output_format_id = output_format

            output_class = output_classes[output_format_id]

            next_decoder = output_class(f, driver, logic_channels=driver.logic_channels, **output_options)
            next_decoder.reset()
            next_decoder.start()
            next_decoder.run(driver)

            next_decoder.stop()



import array
import zipfile
import configparser
import struct
import pathlib
import io

from .output import Output

from . import cond_matches, __version__

TYPECODE = {
    1: "B",
    2: "H",
    4: "L"
}

UNITS = {
    "Hz":  1,
    "kHz": 1000,
    "KHz": 1000,
    "MHz": 1000000,
    "mHz": 1000000,
}

class SrZipInput:
    name = "srzip"
    desc = "srzip session file format data"
    def __init__(self, filename, initial_state=None):
        self.zip = zipfile.ZipFile(filename)
        # self.zip.printdir()
        self.metadata = configparser.ConfigParser()
        self.version = int(self.zip.read("version").decode("ascii"))
        self.metadata.read_string(self.zip.read("metadata").decode("ascii"))
        self.samplenum = -1
        self.matched = None
        self.single_file = "logic-1" in self.zip.namelist()
        # for s in self.metadata.sections():
        #     print(s)
        #     for o in self.metadata.options(s):
        #         print(" ", o, self.metadata.get(s, o))

        samplerate = self.metadata.get("device 1", "samplerate", fallback="0")
        # print(samplerate)
        if " " in samplerate:
            num, units = samplerate.split(" ")
            self.samplerate = float(num) * UNITS[units]
        else:
            self.samplerate = int(samplerate)

        if initial_state:
            self.last_sample = 0
            for channel in initial_state:
                self.last_sample |= initial_state[channel] << channel
        else:
            self.last_sample = None
        self.unitsize = int(self.metadata.get("device 1", "unitsize"))
        self.typecode = TYPECODE[self.unitsize]

        if self.single_file:
            self.data = self.zip.read("logic-1")
            if self.unitsize > 1:
                self.data = array.array(self.typecode, self.data)
        else:
            self.data = None
            self._file_start = -1
            self._file_index = 1

    def wait(self, conds=None):
        self.matched = [False]
        while not any(self.matched):
            self.matched = [True] * len(conds)
            self.samplenum += 1
            if self.single_file:
                if self.samplenum >= len(self.data):
                    raise EOFError()
                sample = self.data[self.samplenum]
            else:
                file_samplenum = self.samplenum - self._file_start
                if self.data is None or file_samplenum >= len(self.data):
                    self._file_start = self.samplenum
                    try:
                        self.data = self.zip.read(f"logic-1-{self._file_index:d}")
                    except KeyError:
                        raise EOFError()

                    if self.unitsize > 1:
                        self.data = array.array(self.typecode, self.data)

                    file_samplenum = 0
                    self._file_index += 1
                sample = self.data[file_samplenum]

            if self.last_sample is None:
                self.last_sample = sample

            for i, cond in enumerate(conds):
                if "skip" in cond:
                    cond["skip"] -= 1
                    self.matched[i] = cond["skip"] == 0
                    continue
                self.matched[i] = cond_matches(cond, self.last_sample, sample)
            self.last_sample = sample

        bits = []
        for b in range(self.unitsize * 8):
            bits.append((sample >> b) & 0x1)

        return tuple(bits)

CHUNK_SIZE = 4 * 1024 * 1024

class SrZipOutput(Output):
    name = "srzip"
    desc = "srzip session file format data"
    def __init__(self, filename, driver, logic_channels=[], analog_channels=[], annotations=[]):
        if annotations:
            raise NotImplementedError("Annotations can't be saved into .sr files.")

        self.zip = zipfile.ZipFile(filename, "w", compression=zipfile.ZIP_DEFLATED)

        self.zip.writestr("version", "2")
        self.metadata = configparser.ConfigParser()
        # libsigrok skips the global section.
        self.metadata.add_section("global")
        self.metadata.set("global", "pysigrok version", __version__)

        self.metadata.add_section("device 1")
        self.metadata.set("device 1", "driver", driver.name)
        self.metadata.set("device 1", "samplerate", str(driver.samplerate))
        self.logic_buffer = None
        if logic_channels:
            self.capturefile = "logic-1"
            self.count = 1
            self.metadata.set("device 1", "capturefile", self.capturefile)
            self.unitsize = len(logic_channels) // 8 + 1
            self.metadata.set("device 1", "unitsize", str(self.unitsize))
            self.metadata.set("device 1", "total probes", str(len(logic_channels)))
            for i, channelname in enumerate(logic_channels):
                self.metadata.set("device 1", f"probe{i+1:d}", channelname)
            self.logic_buffer = array.array(TYPECODE[self.unitsize])
        if analog_channels:
            self.metadata.set("device 1", "total analog", len(analog_channels))
            for i, channelname in enumerate(analog_channels):
                self.metadata.set("device 1", f"analog{i+1:d}", channelname)

        with self.zip.open("metadata", "w") as f:
            self.metadata.write(io.TextIOWrapper(f))


    def decode(self, startsample, endsample, data):
        ptype = data[0]
        if ptype == "logic":
            for _ in range(startsample, endsample):
                self.logic_buffer.append(data[1])
                if len(self.logic_buffer) >= CHUNK_SIZE:
                    fn = f"{self.capturefile}-{self.count:d}"
                    self.zip.writestr(f"{self.capturefile}-{self.count:d}", self.logic_buffer.tobytes())
                    self.count += 1
                    self.logic_buffer = array.array(TYPECODE[self.unitsize])

    def stop(self):
        if self.logic_buffer:
            fn = f"{self.capturefile}-{self.count:d}"
            self.zip.writestr(fn, self.logic_buffer.tobytes())
        self.zip.close()



import click

import functools
import sys

from . import srzip
from . import *

class Output:
    def __init__(self, outfile, output_type, decoder):
        self.outfile = outfile
        self.output_type = output_type
        self.decoder = decoder

    def decode(self, startsample, endsample, data):
        if self.output_type == OUTPUT_PYTHON:
            self.outfile.write(f"{startsample}-{endsample} {self.decoder.id}: {data}\n")

        elif self.output_type == OUTPUT_ANN:
            annotation = self.decoder.annotations[data[0]]
            data = " ".join(["\"" + str(x) + "\"" for x in data[1]])
            self.outfile.write(f"{startsample}-{endsample} {self.decoder.id}: {annotation[0]}: {data}\n")

        elif self.output_type == OUTPUT_BINARY:
            data = " ".join([f"{x:02x}" for x in data[1]])
            self.outfile.write(f"{startsample}-{endsample} {self.decoder.id}: {data}\n")

OUTPUT_TYPES = {
    "python": OUTPUT_PYTHON,
    "exception": OUTPUT_PYTHON,
    "annotation": OUTPUT_ANN,
    "binary": OUTPUT_BINARY,
}

class OrderedParamsCommand(click.Command):
    _options = []

    def parse_args(self, ctx, args):
        parser = self.make_parser(ctx)
        opts, _, param_order = parser.parse_args(args=list(args))
        for param in param_order:
            v = opts[param.name]
            if isinstance(v, list):
                v = v.pop(0)
            type(self)._options.append((param, v))

        return super().parse_args(ctx, args)

@click.command(cls=OrderedParamsCommand)
@click.option('-P', "--protocol-decoder", multiple=True)
@click.option("-p", "--pin-mapping", multiple=True, help="channelname=channelnum")
@click.option("-o", "--channel-option", multiple=True, help="channeloption=value")
@click.option("-N", "--channel-initial-value", multiple=True, help="channelname=initial-pin-value")
@click.option("-i", "--input-file")
@click.option("-O", "--output-format")
@click.option("-f", "--output-file")
def main(protocol_decoder, pin_mapping, channel_option, channel_initial_value, input_file, output_format, output_file):
    decoders = []
    current_decoder = None
    for param, value in OrderedParamsCommand._options:
        if param.name == "protocol_decoder":
            current_decoder = {"id": value, "cls": get_decoder(value), "options": {}, "pin_mapping": {}}
            decoders.append(current_decoder)
        elif param.name == "pin_mapping":
            decoder_id, channelnum = value.split("=")
            channelnum = int(channelnum)
            current_decoder["pin_mapping"][decoder_id] = channelnum
        elif param.name == "channel_option":
            k, v = value.split("=")
            for default_option in current_decoder["cls"].options:
                if default_option["id"] != k:
                    continue
                if isinstance(default_option["default"], int):
                    v = int(v)
            current_decoder["options"][k] = v

    initial_values = {}
    for v in channel_initial_value:
        decoder_id, value = v.split("=")
        value = int(value)
        initial_values[decoders[0]["pin_mapping"][decoder_id]] = value

    if output_file:
        f = open(output_file, "w")
    else:
        f = sys.stdout
    if output_format.count(":") == 1:
        decoder_id, output_name = output_format.split(":")
        output_filter = None
    else:
        decoder_id, output_name, output_filter = output_format.split(":", maxsplit=2)
    output_type = OUTPUT_TYPES[output_name]
    next_decoder = Output(f, output_type, decoders[-1]["cls"])

    data = srzip.SrZipInput(input_file, initial_values)
    all_decoders = []
    for decoder_info in reversed(decoders):
        decoder_class = decoder_info["cls"]
        decoder = decoder_class()
        all_decoders.insert(0, decoder)
        decoder.options = decoder_info["options"]

        # Set any default options that weren't on the command line.
        if hasattr(decoder_class, "options"):
            for option in decoder_class.options:
                if option["id"] not in decoder.options:
                    decoder.options[option["id"]] = option["default"]

        for decoder_id in decoder_info["pin_mapping"]:
            channelnum = decoder_info["pin_mapping"][decoder_id]
            decoder.set_channelnum(decoder_id, channelnum)        

        decoder.add_callback(output_type, output_filter, next_decoder.decode)
        next_decoder = decoder
        output_type = OUTPUT_PYTHON
        output_filter = None

    first_decoder = all_decoders[0]
    first_decoder.input = data
    for d in all_decoders:
        d.reset()
        d.start()
    if data.samplerate > 0:
        first_decoder.metadata(SRD_CONF_SAMPLERATE, data.samplerate)
    
    first_decoder.run(data)

    for d in all_decoders:
        d.stop()

    f.close()



from .output import Output

class BitsOutput(Output):
    name = "bits"
    desc = "ASCII rendering with 0/1"
    def __init__(self, openfile, driver, logic_channels=[], *, width="64"):
        self.width = int(width)
        for i, channel in enumerate(logic_channels):
            print(i, channel)

    def decode(self, startsample, endsample, data):
        ptype = data[0]
        if ptype == "logic":
            print(startsample, endsample, data)



"""Super class for output formats that make them look like decoders."""

class Output:
    def reset(self):
        pass

    def start(self):
        pass

    def run(self, input_):
        channelnum = len(input_.logic_channels)
        any_edge = [{i: "e"} for i in range(channelnum)]
        done = False
        # Store the last value. By searching for the next edge, we're measuring
        # how long the starting value is.
        last_values = None
        while not done:
            startsample = input_.samplenum
            try:
                values = input_.wait(any_edge)
            except EOFError:
                done = True
            endsample = input_.samplenum
            bits = 0
            if last_values:
                for i, v in enumerate(last_values):
                    if v:
                        bits |= 1 << i
                self.decode(startsample, endsample, ["logic", bits])
            last_values = values

    def stop(self):
        pass
